@@ -1,6 +1,6 @@
 #! /usr/bin/perl
 #
-# Copyright (c) 2007-2023, PostgreSQL Global Development Group
+# Copyright (c) 2007-2024, PostgreSQL Global Development Group
 #
 # src/backend/utils/mb/Unicode/UCS_to_GB18030.pl
 #
@@ -14,7 +14,7 @@
 # and the "b" field is the hex byte sequence for UHC
 
 use strict;
-use warnings;
+use warnings FATAL => 'all';
 
 use convutils;
 
